#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import os, sys, time, timeit
if not os.path.dirname(__file__) in sys.path : sys.path.append(os.path.dirname(__file__))
import io, _io
import zlib, zipfile
import typing # typing added in version 3.5, https://docs.python.org/3/library/typing.html

from importlib import abc
from importlib import _bootstrap
from importlib import _bootstrap_external

from tree import Tree

class ZipException(ImportError):
    def __init__(self, *args, **kwargs):
        if False : print(f"[EXP] {args[0]}:::{kwargs}", file=sys.stderr)
        pass

class ZipReader(abc.TraversableResources): # abc.Traversable, abc.TraversableResources
    def __init__(self, loader, module):
        _, _, name = module.rpartition('.')
        self.prefix = loader.virt.replace('\\', '/') + name + '/'
        self.archive = loader.real

    def open_resource(self, resource):
        try:
            return super().open_resource(resource)
        except KeyError as exc:
            raise FileNotFoundError(exc.args[0])

    def is_resource(self, path):
        target = self.files().joinpath(path)
        return target.is_file() and target.exists()

    def files(self):
        path = zipfile.Path(self.archive, self.prefix)
        return path

########################################

SIG_STT_ARCHIVE = b'\x50\x4B\x03\x04'
SIG_END_ARCHIVE = b'\x50\x4B\x05\x06'
CENTRAL_DIR_SZE = 22
MAX_COMMENT_LEN = (1 << 16) - 1

def is_ziparchive(p) -> bool :
    return p.endswith(".z") or p.endswith(".zip")

def is_ziparchive_deep(p) -> bool :
    try:
        with io.open(p, "rb") as z:
            byte = z.read(4)
            return True if byte == SIG_STT_ARCHIVE else False
    except : # is directory
        return False

########################################

USE_CACHED_FILE = False
MAP_CACHED_FILE = dict()
def open(fle) :
    if fle in MAP_CACHED_FILE:
        if False : print(f"[INF] reuse fd {fle}")
        zio = MAP_CACHED_FILE[fle]
    else:
        zio = io.open(fle, 'rb', buffering = 16384) # io.DEFAULT_BUFFER_SIZE : 8192
        MAP_CACHED_FILE[fle] = zio
    return zio

def zipinfo(fle : str) -> tuple:
    try :
        fle = os.path.abspath(fle).replace('\\', '/') # all path must use separator '/'
        stf = _bootstrap_external._path_stat(fle) # get file stat struct
        std = _bootstrap_external._path_stat(os.path.dirname(fle)) # get directory stat struct
    except :
        raise ZipException(f"can't open Zip file: {fle!r}")

    enty = {} # by partname as d/e.txt
    stat = {} # by partname as d/e.txt
    tree = Tree().set(fle)
    try:
        #fle = fle.replace('\\', '/') # all path must use separator '/'
        zio = open(fle) if USE_CACHED_FILE else _io.open_code(fle)

        try:
            zio.seek(-CENTRAL_DIR_SZE, 2)
            header_position = zio.tell()
            buffer = zio.read(CENTRAL_DIR_SZE)
        except OSError:
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
        if len(buffer) != CENTRAL_DIR_SZE:
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
        if buffer[:4] != SIG_END_ARCHIVE:
            try:
                zio.seek(0, 2)
                file_size = zio.tell()
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
            max_comment_start = max(file_size - MAX_COMMENT_LEN - CENTRAL_DIR_SZE, 0)
            try:
                zio.seek(max_comment_start)
                data = zio.read()
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
            pos = data.rfind(SIG_END_ARCHIVE)
            if pos < 0:
                raise ZipException(f'not a Zip file: {fle!r}', path=fle)
            buffer = data[pos:pos + CENTRAL_DIR_SZE]
            if len(buffer) != CENTRAL_DIR_SZE:
                raise ZipException(f"corrupt Zip file: {fle!r}", path=fle)
            header_position = file_size - len(data) + pos

        header_size = _unpack_uint32(buffer[12:16])
        header_offset = _unpack_uint32(buffer[16:20])
        if header_position < header_size:
            raise ZipException(f'bad central directory size: {fle!r}', path=fle)
        if header_position < header_offset:
            raise ZipException(f'bad central directory offset: {fle!r}', path=fle)
        header_position -= header_size
        arc_offset = header_position - header_offset
        if arc_offset < 0:
            raise ZipException(f'bad central directory size or offset: {fle!r}', path=fle)

        try: # start of CD (central directory)
            zio.seek(header_position)
        except OSError:
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
        while True:
            buffer = zio.read(46)
            if len(buffer) < 4:
                raise EOFError('EOF read where not expected')
            # Start of file header
            if buffer[:4] != b'PK\x01\x02':
                break                                # Bad: Central Dir File Header
            if len(buffer) != 46:
                raise EOFError('EOF read where not expected')

            ''''''
            flags        = _unpack_uint16(buffer[8:10])
            compress     = _unpack_uint16(buffer[10:12])
            time         = _unpack_uint16(buffer[12:14])
            date         = _unpack_uint16(buffer[14:16])
            crc          = _unpack_uint32(buffer[16:20])
            data_size    = _unpack_uint32(buffer[20:24])
            file_size    = _unpack_uint32(buffer[24:28])
            name_size    = _unpack_uint16(buffer[28:30])
            extra_size   = _unpack_uint16(buffer[30:32])
            comment_size = _unpack_uint16(buffer[32:34])
            file_offset  = _unpack_uint32(buffer[42:46])
            header_size  = name_size + extra_size + comment_size
            timestamp    = int(datetime(date, time))
            ''''''

            if file_offset > header_offset:
                raise ZipException(f'bad local header offset: {fle!r}', path=fle)
            file_offset += arc_offset

            try:
                name = zio.read(name_size)
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

            if len(name) != name_size:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

            try:# on windows, calling fseek to skip over the fields we don't use is slower than reading the data
                # because fseek flushes stdio's internal buffers.    See issue #8745.
                if len(zio.read(header_size - name_size)) != header_size - name_size:
                    raise ZipException(f"can't read Zip file: {fle!r}", path=fle)
            except OSError:
                raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

            if flags & 0x800:
                name = name.decode()
            else:
                try:
                    name = name.decode('ascii')
                except UnicodeDecodeError:
                    raise ZipException(f"unicode decode error: {fle!r}", path=fle)

            is_d = name.endswith('/')    # is dir
            path = '/'.join([fle, name]) # dir is always endswith '/'
            _nme = [p for p in name.split('/') if 0 < len(p)][-1] # file/dir name only ::: low performance
            if False : print(name + " ::: " + _nme)
            '''
            - name : by partname as d/e.txt
            - path : by fullname as drv:/a/b/c.z/d/e.txt
            '''
            nt = \
            { # entry data
                "isd" : is_d       , # is dir
                "pth" : path       , # drv:/a/b/c.z/d/e.txt
                "ent" : name       , # d/e.txt
                "nme" : _nme       , # e.txt
                "met" : compress   , # compression method refer to https://en.wikipedia.org/wiki/ZIP_(file_format)
                "esz" : data_size  , # encrypt data size
                "dsz" : file_size  , # decrypt data size
                "pos" : file_offset, # data position
                "tme" : timestamp  , # datetime
                "crc" : crc        , # crc-code
            }

            s_ = std if is_d else stf
            st = os.stat_result((s_.st_mode,
                                 s_.st_ino,
                                 s_.st_dev,
                                 s_.st_nlink,
                                 s_.st_uid,
                                 s_.st_gid,
                                 file_size,
                                 timestamp,
                                 timestamp,
                                 timestamp,
                                 0,  # ? s_.st_blocks,
                                 0,  # ? s_.st_blksize,
                                 0,  # ? s_.st_rdev,
                                 0, 0, 0, 0, 0))

            enty[name] = nt  # by partname as d/e.txt
            stat[name] = st  # by partname as d/e.txt
            t = tree.addpath(name, (nt, st))
            nt["_tr"] = t
            nt["_st"] = st
            pass
    except OSError:
        raise ZipException(f"can't open Zip file: {fle!r}")
    finally:
        if USE_CACHED_FILE == False : zio.close()

    if False : tree.debug()

    return (enty, stat, tree)

########################################

def getbytes(fle : str, ent : dict):
    try:
        isd = ent["isd"]  # is dir
        nme = ent["nme"]  # simple name
        esz = ent["esz"]  # encrypt data size
        pos = ent["pos"]  # data position
        met = ent["met"]  # compression method
    except OSError:
        raise ZipException(f"invalid entry: {ent!r}", path=fle)

    if isd:
        if False : print(f"cannot decompress dir entry: {nme!r}", file=sys.stderr)
        return

    try:
        fle = fle.replace('\\', '/') # all path must use separator '/'
        zio = open(fle) if USE_CACHED_FILE else _io.open_code(fle)

        try:
            zio.seek(pos)
        except OSError :
            raise ZipException(f"can't read zip file: {fle!r}", path=fle)

        buf = zio.read(30)
        if len(buf) != 30 :
            raise ZipException(f"eof read where not expected: {fle!r}", path=fle)
        if buf[:4] != SIG_STT_ARCHIVE : # bad: local file header
            raise ZipException(f"bad local file header: {fle!r}", path=fle)

        name_size = _unpack_uint16(buf[26:28])
        extr_size = _unpack_uint16(buf[28:30])
        head_size = 30 + name_size + extr_size
        pos += head_size  # Start of file data

        try:
            zio.seek(pos)
        except OSError :
            raise ZipException(f"can't read Zip file: {fle!r}", path=fle)

        raw = zio.read(esz)
        if len(raw) != esz:
            raise ZipException(f"data size mismatch : {fle!r}", path=fle)
    except OSError:
        raise ZipException(f"can't open Zip file: {fle!r}")
    finally:
        if USE_CACHED_FILE == False : zio.close()

    if esz < 0 :
        raise ZipException(f"invalid data size: {esz!r}", path=fle)

    bin = raw if met == 0 else zlib.decompress(raw, -15)
    return bin # refer to https://en.wikipedia.org/wiki/ZIP_(file_format)

########################################

def _pack_uint32(x):
    return (int(x) & 0xFFFFFFFF).to_bytes(4, 'little')

def _unpack_uint32(data):
    assert len(data) == 4
    return int.from_bytes(data, 'little')

def _unpack_uint16(data):
    assert len(data) == 2
    return int.from_bytes(data, 'little')

def datetime(d, t):
    return time.mktime((
        (d >> 9) + 1980,    # bits 9..15: year
        (d >> 5) & 0xF,     # bits 5..8: month
        d & 0x1F,           # bits 0..4: day
        t >> 11,            # bits 11..15: hours
        (t >> 5) & 0x3F,    # bits 8..10: minutes
        (t & 0x1F) * 2,     # bits 0..7: seconds / 2
        -1, -1, -1))

########################################
import times
def _test01(file : str) :
    ntry, stat, tree = zipinfo(file)
    if False : tree.debug()

if __name__ == "__main__":
    stt = times.current_milli()
    _test01(os.path.join(os.environ["PROJECT_HOME"].replace('\"', ''), "lib.p12/site-packages.transformers-4.52.3.z"))
    _test01(os.path.join(os.environ["PROJECT_HOME"].replace('\"', ''), "lib.p12/site-packages.torch-2.6.0+cu126.z"))
    end = times.current_milli()
    print(f"[INF] elapsed time {(end - stt)} ms ...")
    pass





#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import os, sys
import importlib, inspect, traceback
import typing # typing added in version 3.5, https://docs.python.org/3/library/typing.html

DBG = False

def ismod(path : str) -> bool :
    try:
        m = importlib.import_module(path)
        if m is None : return False
        inspect.ismodule(m)
        return inspect.ismodule(m) #m.__name__ == 'pathlib'
    except Exception as e:
        return False

def isclz(head : str, tail : str) -> bool :
    try:
        m = importlib.import_module(head)
        c = getattr(m, tail)
        return inspect.isclass(c)
    except Exception as e:
        return False

def isfun(module_or_class : object, tail : str) -> bool :
    try:
        if module_or_class is None : return False
        f = getattr(module_or_class, tail)
        return inspect.isfunction(f) or inspect.isbuiltin(f) #True
    except Exception as e:
        return False

def getmod(path : str) -> object :
    try:
        m = importlib.import_module(path)
        return m
    except Exception as e:
        return None

def getclz(head : str, tail : str) -> object :
    try:
        m = importlib.import_module(head)
        c = getattr(m, tail)
        return c
    except Exception as e:
        return None

########################################

def decompose(funcpath : str) -> tuple[object, object, object] : # dotted path : a.b.c -> a.b and c
    if (funcpath is None) : return (None, None, None)
    sep = funcpath.split(".")
    mod = None
    cls = None
    fun = None
    for i in range(len(sep)) :
        try :
            head = '.'.join(sep[:i+1])
            tail = sep[i+1] if i+1 < len(sep) else ''
            cur_is_m = ismod(head)
            if cur_is_m : mod = getmod(head)
            cur_is_c = isclz(head, tail)
            if cur_is_c : cls = getclz(head, tail)
            cur_is_f = isfun(cls if cls is not None else mod, tail)
            if cur_is_f : fun = tail
            if False : print(f"[{i}] [{head}/{tail}] m[{cur_is_m}] c[{cur_is_c}] f[{cur_is_f}]")
        except Exception as e:
            pass

    if DBG :print(f"[INF] [{funcpath}] == m[{'' if mod is None else mod.__name__}] c[{'' if cls is None else cls.__name__}] f[{'' if fun is None else fun}]", file=sys.stdout)
    return (mod, cls, fun)

def set(mod : object, cls : object, fun : str, val : object) -> None :
    if mod is None and cls is None : return
    if cls is not None : setattr(cls, fun, val)
    else               : setattr(mod, fun, val)
    pass

def get(mod : object, cls : object, fun : str) -> object :
    if mod is None and cls is None : return
    if cls is not None : return getattr(cls, fun)
    else               : return getattr(mod, fun)
    pass

if __name__ == '__main__':
    DBG = True
    import zimport
    print('----------')
    mod, cls, fun = decompose("pathlib")
    mod, cls, fun = decompose("pathlib.Path")
    mod, cls, fun = decompose("pathlib.Path.read_text")
    mod, cls, fun = decompose("pathlib.Path.read_bytes")
    mod, cls, fun = decompose("importlib.machinery.FileFinder.find_spec")
    mod, cls, fun = decompose("builtins.open")
    mod, cls, fun = decompose("zimport.util.zip.is_ziparchive_deep")
    mod, cls, fun = decompose("zimport.util.zip.ZipReader.open_resource")
    print('----------')
    pass

'''
----------
[INF] [pathlib] == m[pathlib] c[] f[]
[INF] [pathlib.Path] == m[pathlib] c[Path] f[]
[INF] [pathlib.Path.read_text] == m[pathlib] c[Path] f[read_text]
[INF] [pathlib.Path.read_bytes] == m[pathlib] c[Path] f[read_bytes]
[INF] [importlib.machinery.FileFinder.find_spec] == m[importlib.machinery] c[FileFinder] f[find_spec]
[INF] [builtins.open] == m[builtins] c[] f[open]
[INF] [zimport.util.zip.is_ziparchive_deep] == m[zimport.util.zip] c[] f[is_ziparchive_deep]
[INF] [zimport.util.zip.ZipReader.open_resource] == m[zimport.util.zip] c[ZipReader] f[open_resource]
----------
'''


#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import io, os, sys, time, shutil, builtins
import zipfile
import pathlib
from .util.path import path_exists_native, slashpath, find

DBG = False

########################################

ZIP_IMPORTED_STRING = [".z/", ".zip/", ".z\\", ".zip\\"]
# ZIP_IMPORTED_STRING = [".z/", ".z\\"]
ZIP_IMPORTED_CACHES = ["/.cache/", "\\.cache\\"]
def is_zip_path(path) : # return just true, path == a.zip/a/b/c.x
    if not any (tok in path for tok in ZIP_IMPORTED_STRING) : return False # if not (".z/" in path or ".zip/" in path or ".z\\" in path or ".zip\\" in path) : return False
    if any (tok in path for tok in ZIP_IMPORTED_CACHES) : return False
    return True

def is_cac_path(path) : # return just true, path == a.zip/a/b/c.x
    if not any (tok in path for tok in ZIP_IMPORTED_STRING) : return False
    if any (tok in path for tok in ZIP_IMPORTED_CACHES) : return True
    return False

########################################
def extract(zip_path, zip_cached_path, ent_path) :
    try:
        with zipfile.ZipFile(zip_path) as zip_file:
            zip_file.extract(ent_path, zip_cached_path)
    except FileNotFoundError:
        print(f"[HOOK:::zip][ERROR] zip file not found: {zip_path}", file=sys.stderr)
    except zipfile.BadZipFile:
        print(f"[HOOK:::zip][ERROR] bad zip file: {zip_path}", file=sys.stderr)
    except Exception as e:
        print(f"[HOOK:::zip][ERROR] extracting {ent_path} from {zip_path}: {e}", file=sys.stderr)

def extract_post(bin_path) :
    if not os.path.exists(bin_path) : return
    path_only = bin_path[:bin_path.rfind('/')] # a/b/c/d.exe to a/b/c
    name_only = bin_path[bin_path.rfind('/'):] # a/b/c/d.exe to /d.exe
    if name_only.endswith('.exe') :
        if ('ffmpeg' in name_only) : # case of ffmpeg-win-x86_64-v?.?.exe to ffmpeg.exe
            shutil.copy(bin_path, ''.join([path_only , '/ffmpeg.exe']))
    pass

########################################
def hook_fileio(zimport, name, is_string_path, is_stderr_print, orgfunc) :
    ZIP_NTRY_INFO = zimport.ZIP_NTRY_INFO
    ZIP_STAT_INFO = zimport.ZIP_STAT_INFO
    ZIPCACHED_DIR = zimport.cached_dir

    def funcwithstring(*args, **kwargs):
        claz = type(args[0])
        path = args[0]
        hook = orgfunc
        if claz is not str: return hook(*args, **kwargs)  # type of Path class
        if path.startswith('.'): return hook(*args, **kwargs)  # relative path
        if is_zip_path(path):
            args, ret = bypass(hook, name, *args, **kwargs)
            if DBG : print("[HOOK:z:zip] {} {}, {}".format(name, args, kwargs), file = sys.stderr if is_stderr_print else sys.stdout)
        else:
            ret = hook(*args, **kwargs)
            if DBG : print("[HOOK:x:org] {} {}, {}".format(name, args, kwargs), file = sys.stderr if is_stderr_print else sys.stdout)
        return ret

    def funcwithpath(*args, **kwargs):  # for sklearn
        path = args[0].as_posix()
        hook = orgfunc
        if path.startswith('.'): return hook(*args, **kwargs)  # relative path
        if is_zip_path(path):
            args, ret = bypass(hook, name, *args, **kwargs)
            if DBG : print("[HOOK:z:zip] {} {}, {}".format(name, args, kwargs), file = sys.stderr if is_stderr_print else sys.stdout)
        else:
            ret = hook(*args, **kwargs)
            if DBG : print("[HOOK:x:zip] {} {}, {}".format(name, args, kwargs), file = sys.stderr if is_stderr_print else sys.stdout)
        return ret

    def bypass(hook, name, *args, **kwargs):
        isString = True if type(args[0]) is str else False
        org_path = args[0]
        zip_path, ent_path, new_path = zimport.encache_path(org_path)
        if DBG : print(f"[HOOK:::zip] org_path [{org_path}] == [{zip_path}] + [{ent_path}]")
        if DBG : print(f"[HOOK:::zip] new_path [{new_path}]")
        if not zip_path in ZIP_NTRY_INFO :
            return args, hook(*args, **kwargs)
        zip_list = ZIP_NTRY_INFO[zip_path]
        isfolder = True if (''.join([ent_path, '/'])) in zip_list else False
        if not "stat" in name :
            if not path_exists_native(new_path):  # path_exists_native, os.path.exists or builtins_exists ???
                if isfolder:
                    os.makedirs(new_path, exist_ok=True)
                elif ent_path in zip_list:
                    print("[HOOK:::zip] cache {} [{}]".format(name, new_path), file=sys.stderr)
                    extract(zip_path, ZIPCACHED_DIR(zip_path), ent_path)
                    extract_post(''.join([ZIPCACHED_DIR(zip_path) , '/', ent_path]))
            args = tuple([(new_path if isString else pathlib.Path(new_path)) if idx == 0 else v for idx, v in enumerate(args)])
            return args, hook(*args, **kwargs)
        else :
            if not zip_path in ZIP_STAT_INFO :
                print("[HOOK:::zip] cache error {} [{}]".format(name, zip_path), file=sys.stderr)
                return args, hook(*args, **kwargs)
            entries = ZIP_STAT_INFO[zip_path]
            neopath = (ent_path + "/") if isfolder else ent_path
            if not neopath in entries : # if exists neopath, make a dir/file
                if not path_exists_native(new_path):  # path_exists_native, os.path.exists or builtins_exists ???
                    if isfolder:
                        os.makedirs(new_path, exist_ok=True)
                    elif ent_path in zip_list:
                        print("[HOOK:::zip] cache {} [{}]".format(name, new_path), file=sys.stderr)
                        extract(zip_path, ZIPCACHED_DIR(zip_path), ent_path)
                        extract_post(''.join([ZIPCACHED_DIR(zip_path), '/', ent_path]))
                args = tuple([(new_path if isString else pathlib.Path(new_path)) if idx == 0 else v for idx, v in enumerate(args)])
                return args, hook(*args, **kwargs)
            stat = entries[neopath]
            return args, stat

    hook_func = funcwithstring if is_string_path else funcwithpath
    zimport.register_hook(name, orgfunc, hook_func)
    return hook_func

########################################

import zimport.util.module as module
def detour(zimport, hookname : str) :
    ZIP_NTRY_INFO = zimport.ZIP_NTRY_INFO
    ZIP_STAT_INFO = zimport.ZIP_STAT_INFO
    ZIP_NTRY_TREE = zimport.ZIP_NTRY_TREE
    ZIPCACHED_DIR = zimport.cached_dir

    mod, clz, fun = module.decompose(hookname)
    orgfunc = module.get(mod, clz, fun)

    def hook(*args, **kwargs) :
        def open(p): f = builtins.open(p); n = f.name; f.close(); return n.replace('\\', '/')
        path = slashpath(args[0])
        is_c = is_cac_path(path)
        is_z = is_zip_path(path)

        if (hookname == "tokenize._builtin_open") :
            if False : pass
            elif is_c :
                zip_path, ent_path, new_path = zimport.decache_path(path)
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {new_path}")
                args = tuple([new_path if idx == 0 else v for idx, v in enumerate(args)])
            elif is_z :
                pass
            else :
                pass
            return builtins.open(*args, **kwargs)

        if (hookname == "importlib.machinery.FileFinder.find_spec") : # 20250531 torchvision patch
            if False : pass
            elif is_c :
                zip_path, ent_path, new_path = zimport.decache_path(path)
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {new_path}")
                args[0].path = new_path
            elif is_z :
                zip_path, ent_path, new_path = zimport.encache_path(path)
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {new_path}")
                args[0].path = new_path
            else :
                pass
            return orgfunc(*args, **kwargs)

        if (hookname == "os.path.exists") :
            if False : pass
            elif is_c :
                ret = path_exists_native(path)
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}")
                return ret
            elif is_z :
                zip_path, ent_path, new_path = zimport.encache_path(path)
                if zip_path in ZIP_NTRY_INFO:
                    ret = (ent_path) in ZIP_NTRY_INFO[zip_path] or (ent_path + '/') in ZIP_NTRY_INFO[zip_path]
                    if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}")
                    return ret
                return False
            else :
                pass
            return orgfunc(*args, **kwargs)

        if (hookname == "os.path.isdir") :
            if False : pass
            elif is_c :
                zip_path, ent_path, new_path = zimport.decache_path(path)
                ret = zip_path in ZIP_NTRY_INFO and (ent_path + '/') in ZIP_NTRY_INFO[zip_path]
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}")
                return ret
            elif is_z :
                zip_path, ent_path, new_path = zimport.encache_path(path)
                ret = zip_path in ZIP_NTRY_INFO and (ent_path + '/') in ZIP_NTRY_INFO[zip_path]
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}")
                return ret
            else :
                pass
            return orgfunc(*args, **kwargs)

        if (hookname == "os.path.isfile") : # 20250606 transformers requires this patch !!!
            if False : pass
            elif is_c :
                zip_path, ent_path, new_path = zimport.decache_path(path)
                ret = zip_path in ZIP_NTRY_INFO and (ent_path) in ZIP_NTRY_INFO[zip_path]
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}")
                return ret
            elif is_z :
                zip_path, ent_path, new_path = zimport.encache_path(path)
                ret = zip_path in ZIP_NTRY_INFO and (ent_path) in ZIP_NTRY_INFO[zip_path]
                if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}")
                return ret
            else :
                pass
            return orgfunc(*args, **kwargs)

        if (hookname == "os.listdir") : # 20250602 transformers patch
            if False : pass
            elif is_c :
                zip_path, ent_path, new_path = zimport.decache_path(path)
                if zip_path in ZIP_NTRY_INFO and (ent_path + '/') in ZIP_NTRY_INFO[zip_path]:
                    t = ZIP_NTRY_INFO[zip_path][ent_path + '/']['_tr']
                    ret = list(t.dict().keys())
                    ret.sort()
                    if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}", file=sys.stdout)
                    return ret
                pass
            elif is_z :
                zip_path, ent_path, new_path = zimport.encache_path(path)
                if zip_path in ZIP_NTRY_INFO and (ent_path + '/') in ZIP_NTRY_INFO[zip_path]:
                    t = ZIP_NTRY_INFO[zip_path][ent_path + '/']['_tr']
                    ret = list(t.dict().keys())
                    ret.sort()
                    if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {ret}", file=sys.stdout)
                    return ret
                pass
            else :
                pass
            return orgfunc(*args, **kwargs)

        if (hookname == "os.path.join") :
            if False : pass
            elif is_c :
                return orgfunc(*args, **kwargs)
            elif is_z :
                ret = orgfunc(*args, **kwargs)
                if (ret.endswith(".dll") or ret.endswith(".so") or (".so." in ret) or ret.endswith(".dylib") or ret.endswith(".exe")) :
                    zip_path, ent_path, new_path = zimport.encache_path(ret)
                    if DBG : print(f"[INF:::detour] [{hookname}] ::: {ret} -> {new_path}")
                    return new_path
                return ret
            else :
                pass
            return orgfunc(*args, **kwargs)

        if (hookname == "os.path.dirname") :
            if False : pass
            elif is_c :
                # zip_path, ent_path, new_path = zimport.decache_path(path)
                # if zip_path in ZIP_NTRY_INFO and ((ent_path) in ZIP_NTRY_INFO[zip_path] or (ent_path + '/') in ZIP_NTRY_INFO[zip_path]):
                #     args = tuple([new_path if idx == 0 else v for idx, v in enumerate(args)])
                # if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {args}")
                return orgfunc(*args, **kwargs)
            elif is_z :
                # zip_path, ent_path, new_path = zimport.encache_path(path)
                # if zip_path in ZIP_NTRY_INFO and ((ent_path) in ZIP_NTRY_INFO[zip_path] or (ent_path + '/') in ZIP_NTRY_INFO[zip_path]):
                #     args = tuple([new_path if idx == 0 else v for idx, v in enumerate(args)])
                # if DBG : print(f"[INF:::detour] [{hookname}] ::: {path} -> {args}")
                return orgfunc(*args, **kwargs)
            else :
                pass

        ret = orgfunc(*args, **kwargs)
        return ret

    module.set(mod, clz, fun, hook)
    zimport.register_hook(hookname, orgfunc, hook)
    return hook




#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------

DBG = False

import os, sys, time, timeit
import io, _io, _imp
import marshal
# import _frozen_importlib as _bootstrap
# import _frozen_importlib_external as _bootstrap_external
from importlib import _bootstrap
from importlib import _bootstrap_external
from .zip import getbytes, _unpack_uint32
from typing import Optional, Tuple # typing added in version 3.5, https://docs.python.org/3/library/typing.html

def compile_from_py(fullpath, src) : # fullpath : a/b/c.z/d/e.py, src : plain-text
    try:
        src = src.replace(b'\r\n', b'\n').replace(b'\r', b'\n')  # _normalize_line_endings(source)
        bin = compile(src, fullpath, 'exec', dont_inherit=True)
        return bin
    except Exception as e:
        if True: print(f"[ERR:::compile] [{e}] in [{fullpath}] ...", file=sys.stderr)
        return None

NUMBA_CHK_FOR_RECOMPILE = True
ALL_PY_FORCED_RECOMPILE = False

def unmarshal_from_pyc(self, fullpath, virtpath, virtname, data): # fullpath : a/b/c.z/d/e.pyc, virtpath : d/e.pyc, virtname d.e
    # Given the contents of a .py[co] file, unmarshal the data and return the code object.
    # Raises ImportError it the magic word doesn't match, or if the recorded .py[co] metadata does not match the source.
    exc_details = { 'name': virtname, 'path': virtpath, }
    flags = _bootstrap_external._classify_pyc(data, virtname, exc_details)
    hash_based = flags & 0b1 != 0
    if hash_based:
        check_source = flags & 0b10 != 0
        if (_imp.check_hash_based_pycs != 'never' and (check_source or _imp.check_hash_based_pycs == 'always')):
            source_bytes = get_source_code_by_pyc(self, virtpath)
            if source_bytes is not None:
                source_hash = _imp.source_hash(_bootstrap_external._RAW_MAGIC_NUMBER, source_bytes,)
                _bootstrap_external._validate_hash_pyc(data, source_hash, virtname, exc_details)
    else:
        if (ALL_PY_FORCED_RECOMPILE) : return None
        source_bytes = get_source_code_by_pyc(self, virtpath)
        
        if (NUMBA_CHK_FOR_RECOMPILE) and (b" numba" in source_bytes) and (not "numba" in virtname) : #14mhz numba related py must be alwarys recompile ...
            if DBG : print(f"[INF:::pathfinder] forced recompile {virtpath}:::{virtname}:::{fullpath}", file=sys.stderr)
            return None

        py_time, py_size =  get_time_and_size_of_py(self, virtpath)
        if py_time : # pyc time vs zip entry time
            pyc_time = _unpack_uint32(data[8:12])
            pyc_size = _unpack_uint32(data[12:16])
            if (pyc_size != py_size) or 2 < abs(pyc_time - py_time) : # lenient timestamp check
                return None
            else :
                pass

    code = marshal.loads(data[16:])
    if not isinstance(code, UNMARSHAL_CODE_TYPE):
        raise TypeError(f'compiled module {fullpath!r} is not a code object')
    return code
UNMARSHAL_CODE_TYPE = type(unmarshal_from_pyc.__code__)

########################################

def get_source_code_by_pyc(self, path) -> Optional[bytes] :
    #assert path[-1:] in ('c', 'o')  # strip 'c' or 'o' from *.py[co]
    if not path or path[-1:] not in ('c', 'o'):
        return None
    path = path[:-1] # pyc, pyo to py
    if (path in self.zent) :
        ntry = self.zent[path]
        data = getbytes(self.real, ntry)
        return data
    else :
        return None

def get_time_and_size_of_py(self, path) -> Tuple[int, int] :
    #assert path[-1:] in ('c', 'o')  # strip 'c' or 'o' from *.py[co]
    if not path or path[-1:] not in ('c', 'o'):
        return None
    path = path[:-1] # pyc, pyo to py
    if (path in self.zent) :
        ntry = self.zent[path]  # entry
        time = int(ntry["tme"]) # timestamp
        size = int(ntry["dsz"]) # decrypt size
        return time, size
    else :
        return 0, 0


#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import os, sys
if not os.path.dirname(__file__) in sys.path : sys.path.append(os.path.dirname(__file__))
import typing # typing added in version 3.5, https://docs.python.org/3/library/typing.html
from typing import Self
from typing import overload

DBG = False
ROOT_SIGNATURE = '/'
class Tree :
    def __init__(self, name : str = None, data = None):
        if name is not None and '/' in name :
            raise Exception(f"[ERR] name can't contains '/' [{name}]")
        self._supr = None
        self._name = name if name is not None else ''
        self._dict = dict()
        self._data = data

    def root(self) -> Self :
        if self.supr() is None : return self
        return self.supr().root()

    def supr(self) : return self._supr
    def name(self) : return self._name
    def dict(self) : return self._dict
    def get(self) : return self._data
    def set(self, o) : self._data = o; return self

    def _nomalized(self, name : str) -> str : # remove ROOT_SIGNATURE or last '/'
        if name.startswith(ROOT_SIGNATURE): name = name[1:]
        if name.endswith('/') : name = name[:-1]
        return name

    def add(self, name : str, data = None) -> Self : # add only one tree item
        if name is None :
            print(f"[ERR] tree is None ...", file=sys.stderr)
            return None
        name = self._nomalized(name)
        if (name in self.dict().keys()) :
            if DBG : print(f"[INF] tree [{self.path()}] already has sub-tree [{name}] ...", file=sys.stderr)
            return self.dict()[name]
        return self.addtree(Tree(name, data))

    def addpath(self, path : str, data = None) -> Self : # add path hierarchy tree item ie, a/b/c/d
        if path is None or len(path) == 0 : return None
        paths = self._nomalized(path).split('/')
        return _addpath(self, paths, data)

    def addtree(self, tree : Self) -> Self : # add only one tree item
        if tree is None :
            print(f"[ERR] tree is None ...", file=sys.stderr)
            return None
        if (tree.name() in self.dict().keys()) :
            if DBG : print(f"[INF] tree [{self.path()}] already has sub-tree [{tree.name()}] ...", file=sys.stderr)
            return self.dict()[tree.name()]
        tree._supr = self
        self._dict[tree.name()] = tree
        return tree

    def remove(self, name : str) -> None :
        if path is None : return
        name = self._nomalized(name)
        if not name in self.dict().keys() : return
        self.dict().pop(name)
        pass

    def find(self, path : str) -> Self : # find full path ie, /a/b/c or a/b/c
        if path is None or len(path) == 0 : return None
        paths = self._nomalized(path).split('/')
        return _find(self, paths)

    def path(self, subpath : str = None) -> str:
        return path(self, subpath)

    def debug(self) -> None :
        debug(self, 0, False)

    def debug_detail(self) -> None :
        debug(self, 0, True)

########################################

def _addpath(tree : Tree, paths : list, data = None, depth : int = 0) -> Tree : # add path hierarchy tree item ie, a/b/c/d
    name = paths[depth]
    if (name in tree.dict().keys()) :
        return _addpath(tree.dict()[name], paths, data, depth+1) if depth+1 < len(paths) else tree.dict()[name]
    else :
        t = tree.add(name) if depth+1 < len(paths) else tree.add(name, data)
        return _addpath(t, paths, data, depth+1) if depth+1 < len(paths) else t
    pass

def _find(tree : Tree, paths : list, depth : int = 0) -> Tree :
    name = paths[depth]
    if (name in tree.dict().keys()) :
        t = tree.dict()[name]
        return _find(t, paths, depth+1) if depth+1 < len(paths) else t
    else :
        return None
    pass

def path(tree : Tree, subpath : str = None) -> str:
    if tree._supr is None :
        return tree.name() + ('' if subpath is None else ROOT_SIGNATURE + subpath)
    else :
        return tree.supr().path(tree.name() + ('' if subpath is None else '/' + subpath))

def debug(tree : Tree, dep : int = 0, detail_display : bool = False) :
    if (0 < dep) :
        if detail_display :
            print(f"{space(dep)}{tree.name()} ({'' if tree.get() is None else tree.get()}) ::: [{tree.path()}]", file=sys.stdout)
        else      :
            print(f"{space(dep)}{tree.name()} ::: [{tree.path()}]", file=sys.stdout)
    for k, v in tree.dict().items():
        debug(v, dep + 1, detail_display)

def space(dep) :
    s = ''
    dep = dep - 1
    for i in range(dep) : s += '    ' if i+1 < dep else '  + '
    return s

########################################

def _test01() :
    print("------------")
    t = Tree()
    t.add('a', 'a').add('b', 'b')
    t.find('/a/b').add('/c').add('1', 1)
    t.find('/a/').find('/b').add('c').add('2', 2)
    t.find('/a/').find('b/').find('c/').add('3', 3)
    t.find('a/b/c').add('4', 4)
    t.find('a/b/').add('c', 'c')
    t.find('a/b/').find('/c/').set('c')
    t.debug()
    print("------------")
    t.find('/a/b/').remove('/c/')
    t.find('/a/').remove('b/')
    t.debug()
    print("------------")
    t.addpath('w/x/y').addpath('z', 'value')
    t.addpath('w/x/y').addpath('o', 'value')
    t.debug()
    pass

def _test02(file : str) :
    import zip as ZIP
    ntry, stat, tree = ZIP.zipinfo(file)
    if False : tree.debug()

import times
if __name__ == '__main__':
    stt = times.current_milli()
    _test01()
    #_test02(os.path.join(os.environ["PROJECT_HOME"].replace('\"', ''), "lib.p12/site-packages.transformers-4.52.3.z"))
    #_test02(os.path.join(os.environ["PROJECT_HOME"].replace('\"', ''), "lib.p12/site-packages.torch-2.6.0+cu126.z"))
    end = times.current_milli()
    print(f"[INF] elapsed time {(end - stt)} ms ...")
    pass




#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import io, os, sys, importlib, time
import traceback
import zimport.util.zip as ZIP
import zimport.util.bootstrap as BOOTSTRAP
import zimport.util.path as PATH

import types
import builtins

from zimport import pathfinder_impl

# from importlib import _bootstrap
# from importlib import _bootstrap_external

from zimport import main

DBG = False
# ENTRIES_BY_ZIP = dict()

class PathFinder(): #_bootstrap_external._LoaderBasics/LoaderBasics
    def __init__(self, path):
        path = os.path.abspath(path) # a/b/./c to a/b/c, a/../a/./b/c to a/b/c #20250529
        path = path.replace('\\', '/') if path is not None else None
        real, virt = PATH.virtual_path_split(path)
        if not isinstance(path, str) :
            raise ZIP.ZipException(f"zip loader exception", path=path)
        if not path :
            raise ZIP.ZipException(f"zip loader exception", path=path)
        if not ZIP.is_ziparchive(real) :
            raise ZIP.ZipException(f"zip loader exception", path=path)
        if '.cache' in path :
            raise ZIP.ZipException(f"zip loader exception", path=path)

        zimport = main.getInstance()
        zimport.invalidate_caches()
        self.real = real                       # real file in filesystem, generally .zip/.z
        self.virt = virt + '/' if virt else '' # virt path is always directory not file
        self.zent = zimport.getentries(real)   # zip entries with 'real' zip path key
        if DBG : print(f"[INF] new loader {real} ::: {virt}")

    def find_spec(self, fullname, target=None):
        has = pathfinder_impl.has_module_is_packge(self, fullname)

        if has is None: # None : not found
            mod_path = pathfinder_impl.get_path_for_module(self, fullname)
            if (mod_path + '/') in self.zent : # this path represent a directory?
                path = f'{self.real}/{mod_path}'
                spec = BOOTSTRAP.modulespec(name=fullname, loader=None, is_package=True)
                spec.submodule_search_locations.append(path) # invoke new loader(path)
            else: # is package or not
                spec = None # the module cannot be found.
        else:
            mod_path = pathfinder_impl.get_module_filename(self, fullname)  # 14mhz
            if mod_path.endswith(".pyd") :
                spec = BOOTSTRAP.modulespec(fullname, self, origin=mod_path, is_package=False)
            else :
                spec = BOOTSTRAP.spec_from_loader(fullname, self, is_package=has)  # invoke get_filename

        if DBG and spec : print(f"[INF:::pathfinder] find_spec : {self.real.rpartition('/')[2]}:::{self.virt}:::{fullname}:::{spec}")
        return spec

    ########################################

    def get_filename(self, fullname) :
        mod_path = pathfinder_impl.get_module_filename(self, fullname) #14mhz
        if DBG : print(f"[INF:::pathfinder] get_filename : {self.real.rpartition('/')[2]}:::{self.virt}:::{fullname} -> {mod_path.rpartition('/')[2]}")
        return mod_path

    def get_code(self, fullname) :
        code, ispackage, mod_path = pathfinder_impl.get_module_code(self, fullname)
        if DBG : print(f"[INF:::pathfinder] get_code : {self.real.rpartition('/')[2]}:::{self.virt}:::{fullname} -> {mod_path.rpartition('/')[2]}")
        return code, ispackage, mod_path

    def get_data(self, pathname):
        data, mod_path = pathfinder_impl.get_data(self, pathname)
        if DBG : print(f"[INF:::pathfinder] get_data : {self.real.rpartition('/')[2]}:::{self.virt}:::{pathname} -> {mod_path.rpartition('/')[2]}")
        return data

    def get_source(self, fullname) : # the source code for the specified module
        src, mod_path = pathfinder_impl.get_source(self, fullname)
        if DBG : print(f"[INF:::pathfinder] get_source : {self.real.rpartition('/')[2]}:::{self.virt}:::{fullname} -> {mod_path.rpartition('/')[2]}")
        return src

    def get_resource_reader(self, fullname) :
        reader = pathfinder_impl.get_resource_reader(self, fullname)
        if DBG : print(f"[INF:::pathfinder] get_resource_reader : {self.real.rpartition('/')[2]}:::{self.virt}:::{fullname}")
        return reader

    ########################################

    def is_package(self, fullname) :
        pass

    def create_module(self, spec) :
        module = types.ModuleType(spec.name)
        module.__spec__ = spec
        return module

    def load_module(self, fullname): #deprecated method
        pass

    def exec_module(self, module):
        def path(p): f = builtins.open(p); n = f.name; f.close(); return n.replace('\\', '/')
        spec = module.__spec__

        if spec is None or spec.origin is None :
            if DBG: print(f"[INF:::exec_module@pathfinder] err : spec is null {[module, ]}", file=sys.stderr)
            pass    

        if DBG: print(f"[INF:::exec_module@pathfinder] exec_module : {spec.name} ::: {spec.origin}")

        if spec.origin.endswith("__init__.py") : # 20240930 pyworld patch load 'package/package.cp312-win_amd64.pyd'
            if DBG: print(f"[INF:::exec_module@pathfinder] exec_module : __init__ [{spec.origin}]", file=sys.stdout)
            pydname = '/'.join([spec.name, spec.name.rpartition('.')[2] + pathfinder_impl._PYTHON__PY_DLL_]) # 'name/name.cp312-win_amd64.pyd'
            if pydname in self.zent :
                replace_pyd = '/'.join([self.real, pydname])
                replace_pth = path(replace_pyd)
                if DBG : print(f"[INF:::exec_module@pathfinder] exec_module : load [{replace_pyd}] to [{replace_pth}]", file=sys.stdout)
                m = self.custom_load_dynamic(module.__name__, replace_pth)

        if spec.origin.endswith(".pyd") or spec.origin.endswith(".so") or (".so." in spec.origin) or spec.origin.endswith(".dylib") : # 20250514 linux patch
            #def path(p): f = builtins.open(p); n = f.name; f.close(); return n.replace('\\', '/')
            name = path(spec.origin)
            #import imp
            #m = imp.load_dynamic(module.__name__, name) # imp deprecated since version 3.4, removed in version 3.12.
            m = self.custom_load_dynamic(module.__name__, name)
            if DBG : print(f"[INF:::exec_module@pathfinder] load {[name,]}", file=sys.stderr)
            return m
        else :
            code, ispackage, mod_path = self.get_code(module.__name__)
            if code is None: raise ImportError('cannot load module {!r} when get_code() returns None'.format(module.__name__))
            spec._cached = mod_path
            #exec(code, module.__dict__)  # mainly exec works here ...
            try:
                exec(code, module.__dict__) # mainly exec works here ...
            except Exception as e:
                if DBG : print(f"[INF:::exec_module@pathfinder] failed [{e}] with [{mod_path}] ...", file=sys.stderr)
                raise ImportError("[ERR] failed import ...")
                pass

    def load_module(self, fullname):
        return BOOTSTRAP.load_module_shim(self, fullname)

    #https://stackoverflow.com/questions/60158932/how-to-replace-usages-of-deprecated-imp-load-dynamic
    #https://stackoverflow.com/questions/24166080/importing-dll-into-python-3-without-imp-load-dynamic
    #https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    def custom_load_dynamic(self, path, fullname):
        if DBG : print(f"[INF:::custom_load_dynamic@pathfinder] self[{self}] path[{path}] file[{fullname}]")
        spc = importlib.util.spec_from_file_location(path, fullname)
        mod = importlib.util.module_from_spec(spc)
        sys.modules[path] = mod
        spc.loader.exec_module(mod)
        return mod

    ########################################

    def invalidate_caches(self) :
        try:
            zimport = main.getInstance()
            zimport.fixarchive(self.real)
            self.zent = zimport.getentries(self.real)
        except :
            self.zent = {}

    def __repr__(self):
        return f'<zimporter object "{self.real}/{self.virt}">'

    ########################################




#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------

import os, sys

__version__ = '0.1.9'
__all__ = ["main"]

if sys.version_info < (3,9):
    raise Exception("[ERR] zimport requires Python 3.9 or greater ...")

from .main import install, uninstall, debug, zimport_set_cache_dir, zimport_clear_cache, zimport_extract_to_cache
install()
debug(False)



#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import os, sys, time, timeit
import io, _io
# import _frozen_importlib as _bootstrap
# import _frozen_importlib_external as _bootstrap_external
from importlib import _bootstrap
from importlib import _bootstrap_external
import pathlib
import _frozen_importlib_external

def slashpath(p : object) -> str:
    if False : pass
    elif (type(p) is str) : pass
    elif (type(p) is pathlib.PosixPath) :
        p = p.as_posix()
    elif (type(p) is pathlib.WindowsPath) :
        p = p.as_posix()
    elif (type(p) is _frozen_importlib_external.FileFinder) : p = p.path
    else :
        print(f"--------------------------------------------------------- must be check", file=sys.stderr)

    a = os.path.abspath(p)
    u = a.replace('\\', '/') # to unix '/' path
    return u

def path_exists_native(p) : # exists directory or file
    try : # refer to https://happytest-apidoc.readthedocs.io/en/latest/_modules/_frozen_importlib_external/
        stat = _bootstrap_external._path_stat(p)
        return True
    except :
        return False

def path_split(path):
    head, _, tail = path.rpartition('/')
    return head, tail

def path_join(*path_parts):
    return '/'.join([part.rstrip('/') for part in path_parts if part])

def virtual_path_split(p) : #
    # for benchmark, https://stackoverflow.com/questions/28859095/most-efficient-method-to-check-if-dictionary-key-exists-and-process-its-value-if
    if p is None or len(p) == 0 : print(f"[ERR] invalid path {p!r}");  return '', ''
    real = p.rstrip('/')  # real file in filesystem, generally .zip/.z
    virt = []             # virtual path, dir/file
    while True :
        if path_exists_native(real) :
            return real, '/'.join(virt) # ex, a/b/c.z/d to a/b/c.z and d
        else :
            dir, _, sub = real.rpartition('/')
            real = dir
            virt.insert(0, sub)
            if len(real) == 0 : return real, '/'.join(virt) # error! ie, a/b/c.z/d to '' and a/b/c.z/d

def find(path) :
    dir = None
    if   os.path.exists(__file__ + "/./" + path)              : dir = os.path.abspath(__file__ + "/./" + path).replace('\\', '/')
    elif os.path.exists(__file__ + "/../" + path)             : dir = os.path.abspath(__file__ + "/../" + path).replace('\\', '/')
    elif os.path.exists(__file__ + "/../../" + path)          : dir = os.path.abspath(__file__ + "/../../" + path).replace('\\', '/')
    elif os.path.exists(__file__ + "/../../../" + path)       : dir = os.path.abspath(__file__ + "/../../../" + path).replace('\\', '/')
    elif os.path.exists(__file__ + "/../../../../" + path)    : dir = os.path.abspath(__file__ + "/../../../../" + path).replace('\\', '/')
    elif os.path.exists(__file__ + "/../../../../../" + path) : dir = os.path.abspath(__file__ + "/../../../../../" + path).replace('\\', '/')
    return dir

def exists(base : str, dirs : list, name : str) :
    for d in dirs :
        p = os.path.abspath('/'.join([base, d, name])).replace('\\', '/')
        if os.path.exists(p) :
            return p
    return None



#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import os, sys, importlib

import zimport.util.path as PATH
import zimport.util.zip as ZIP
import zimport.util.compile as COMPILE

DBG = False

#_suffixes = [s[0] for s in imp.get_suffixes() if s[2] == imp.C_EXTENSION] # imp deprecated since version 3.4, removed in version 3.12.
_suffixes = importlib.machinery.EXTENSION_SUFFIXES
#print(f"_suffixes ::: {_suffixes}") # ex) _suffixes ::: ['.cp311-win_amd64.pyd', '.pyd']
_PYTHON_VERSION_ = f"{sys.version_info.major}{sys.version_info.minor}"
_PYCACHE_SUFFIX_ = ".cpython-" + _PYTHON_VERSION_ + ".pyc"
#_PYTHON__PY_DLL_ = [s[0] for s in imp.get_suffixes() if s[2] == imp.C_EXTENSION][0]  # imp deprecated since version 3.4, removed in version 3.12.
_PYTHON__PY_DLL_ = _suffixes[0]
#print(f"_PYTHON__PY_DLL_ ::: {_PYTHON__PY_DLL_}") # ex) _PYTHON__PY_DLL_ ::: .cp311-win_amd64.pyd
_zip_searchorder = \
(
    ('/__init__.pyc', True, True, False),
    ('/__init__.py', False, True, False),
    #('/__init__.pyi', False, True, False), # Do not import, refer PEP-484 document
    ('.pyc', True, False, False),
    ('.py', False, False, False),
    #('.pyi', False, False, False),         # Do not import, refer PEP-484 document
    ('.pyd', False, False, True),
    (_PYTHON__PY_DLL_, False, False, True),
) # suffix, isbytecode, ispackage, isdll

def get_path_for_module(self, fullname):
    return self.virt + fullname.rpartition('.')[2]

def has_module_is_packge(self, fullname):
    path = get_path_for_module(self, fullname)
    for suffix, isbytecode, ispackage, isdll  in _zip_searchorder: #14mhz
        fullpath = path + suffix
        if fullpath in self.zent:
            return ispackage
    return None # cannot found in current zip file

####

def get_module_filename(self, fullname):
    path = get_path_for_module(self, fullname)
    for suffix, isbytecode, ispackage, isdll in _zip_searchorder:
        fullpath = path + suffix
        if fullpath in self.zent :
            toc_entry = self.zent[fullpath]
            modpath = toc_entry["pth"]
            return modpath
        else :
            continue
    if DBG: print(f"[:path_hook] can't find module => {fullname!r}", file=sys.stderr)
    return None

####

def get_module_code(self, fullname : str):
    path = get_path_for_module(self, fullname)
    for suffix, isbytecode, ispackage, isdll  in _zip_searchorder: #14mhz
        fullpath = path + suffix
        pycached = '' # a/b.pyc -> a/__pycache__/b.cpython-311.pyc'
        if isbytecode :
            if ispackage : # '__init__' series
                new_suffix = "__init__" + _PYCACHE_SUFFIX_ # ex) __init__.cpython-311.pyc
                cachedtemp = path.split('/')
                cachedtemp.append("__pycache__")
                cachedtemp.append(new_suffix)
                pycached = ('/').join(cachedtemp)
            else :
                new_suffix =_PYCACHE_SUFFIX_
                cachedtemp = (path + new_suffix).split('/')
                cachedtemp.insert(len(cachedtemp) - 1, "__pycache__")
                pycached = ('/').join(cachedtemp)

        #if suffix == '.pyc' : continue # test code
        #if suffix == '.pyc' and path.endswith('config') : continue # test code

        if pycached in self.zent :
            ntry = self.zent[pycached]
        elif fullpath in self.zent :
            ntry = self.zent[fullpath]
        else : continue

        mpth = ntry["pth"]
        data = ZIP.getbytes(self.real, ntry)
        code = None

        if isbytecode:
            code = COMPILE.unmarshal_from_pyc(self, mpth, fullpath, fullname, data)
            if False : print(f"[:path_hook] unmarshal => {fullpath}", file=sys.stderr)
        else:
            code = COMPILE.compile_from_py(mpth, data)
            if DBG : print(f"[:path_hook] compile => {fullpath}", file=sys.stderr)

        if code is None: # bad magic number or non-matching mtime
            continue
        return code, ispackage, mpth

    if DBG : print(f"[:path_hook] can't find module => {fullname!r}", file=sys.stderr)
    return None

####

def get_data(self, pathname):
    path = pathname.replace('\\', '/')
    fullpath = path[len(self.real + '/'):] if path.startswith(self.real + '/') else path
    if (fullpath in self.zent) :
        ntry = self.zent[fullpath]
        modpath = ntry["pth"]
        data = ZIP.getbytes(self.real, ntry)
        return data, modpath
    else :
        raise OSError(0, '', pathname)

####

def get_source(self, fullname):# the source code for the specified module
    # print("get_src - " + fullname)
    has = has_module_is_packge(self, fullname)
    if has is None : # None : not found
        return None, None
    else : # is package or not
        mod_path = get_path_for_module(self, fullname)
        fullpath = PATH.path_join(mod_path, '__init__.py').replace('\\', '/') if has else f'{mod_path}.py'

        if (fullpath in self.zent) :
            toc_entry = self.zent[fullpath]
            mod_path = toc_entry["pth"]
            src = ZIP.getbytes(self.real, toc_entry).decode()
            return src, mod_path
        else :
            return None, None

####

def get_resource_reader(self, fullname):
    try:
        if has_module_is_packge(self, fullname):
            return ZIP.ZipReader(self, fullname)
    except :
        return None
####



#-------------------------------------------------------------------------------
# zimport v0.1.9 20250608
# by 14mhz@hanmail.net, zookim@waveware.co.kr
#
# This code is in the public domain
#-------------------------------------------------------------------------------
import os, sys, shutil
if not os.path.dirname(__file__) in sys.path : sys.path.append(os.path.dirname(__file__))
import typing # typing added in version 3.5, https://docs.python.org/3/library/typing.html

from path import exists, slashpath, path_exists_native

CACHE_DIR_NAME = '.cache'
CACHE_DIR_ROOT = None

def init_cached_dir() :
    global CACHE_DIR_ROOT
    if "PROJECT_HOME" in os.environ :
        FROM_STRING = "os.environ['PROJECT_HOME']"
        os.environ["PROJECT_HOME"] = os.environ["PROJECT_HOME"].replace('\"', '') # remove '"'
        CACHE_DIR_ROOT = exists(os.environ["PROJECT_HOME"], ['.', '..', 'lib', 'lib/site-packages'], CACHE_DIR_NAME)
        if CACHE_DIR_ROOT is None : CACHE_DIR_ROOT = os.environ["PROJECT_HOME"] + '/' + CACHE_DIR_NAME
    else :
        FROM_STRING = "os.path.dirname(sys.executable)"
        CACHE_DIR_ROOT = exists(os.path.dirname(sys.executable), ['.', '..', 'lib', 'lib/site-packages'], CACHE_DIR_NAME)
        if CACHE_DIR_ROOT is None : CACHE_DIR_ROOT = os.path.dirname(sys.executable) + '/' + CACHE_DIR_NAME

    CACHE_DIR_ROOT = slashpath(CACHE_DIR_ROOT)
    if not path_exists_native(CACHE_DIR_ROOT) : os.makedirs(CACHE_DIR_ROOT, exist_ok=True)
    print(f"[INF] zimport cache_dir ::: [{CACHE_DIR_ROOT}] from {FROM_STRING}", file=sys.stderr)

def set_cached_dir(path) :
    global CACHE_DIR_ROOT
    path = slashpath(path)
    if not CACHE_DIR_NAME in path :
        print(f"[INF:::cache] name of the path should be the /{CACHE_DIR_NAME} as the last path. ...", file=sys.stderr)
        return
    if not path_exists_native(path) : os.makedirs(path, exist_ok=True)
    CACHE_DIR_ROOT = path
    pass

def del_cached_dir() :
    if not path_exists_native(CACHE_DIR_ROOT) : return
    for f in os.listdir(CACHE_DIR_ROOT):
        try:
            path = os.path.join(CACHE_DIR_ROOT, f)
            shutil.rmtree(path)
        except Exception :
            pass

def get_cached_dir(path) : # a/b/c.z/d to library.dir/.cache/a/b/c.z/d
    if path is None : print(f"[ERR:::cache] invalid get_cached_dir [{path}] ... ", file=sys.stderr); return None
    z = path[path.rfind('/'):]               # drv:/a/b/c.z to /c.z
    if CACHE_DIR_ROOT :
        return slashpath(''.join([CACHE_DIR_ROOT, z]))
    else :
        return slashpath(''.join([path, '/../', CACHE_DIR_NAME, z]))

########################################
bankofpath = dict()
def clear_bank() -> None :
    bankofpath.clear()

def encache_path(ziparchive, path : str) -> tuple: # convert to en-cache path
    if path in bankofpath: return bankofpath[path]
    unxpath = slashpath(path)     # to drv:/1/2/3/x.zip/a/b/c
    zippath = None                # drv:/1/2/3/x.zip
    entpath = None                # a/b/c
    newpath = None                # to drv:/work/.cache/x.zip/a/b/c
    for p in ziparchive:
        if unxpath.startswith(p): # x.zip/a/b/c.x startswith a.zip
            zippath = p
            entpath = unxpath.replace(''.join([p, '/']), '') if unxpath != p else ''
            break
    newpath = None if zippath is None else ('/' if entpath != '' else '').join([get_cached_dir(zippath), entpath])
    if newpath is None : print(f"[ERR:::cache] invalid encache path [{path}] ... ", file=sys.stderr); return None, None, path
    bankofpath[path] = zippath, entpath, newpath  # save cache
    return zippath, entpath, newpath

def decache_path(ziparchive, path) -> tuple : # convert to de-cache path
    if path in bankofpath: return bankofpath[path]
    unxpath = slashpath(path)                                                   # to drv:/work/.cache/x.zip/a/b/c
    tmppath = unxpath[unxpath.rfind(CACHE_DIR_NAME) + len(CACHE_DIR_NAME) + 1:] # to x.zip/a/b/c
    z, _, e = tmppath.partition('/')  # to x.zip + / + a/b/c
    zippath = None                    # drv:/1/2/3/x.zip
    entpath = e                       # a/b/c
    newpath = None                    # drv:/1/2/3/x.zip + / + a/b/c
    for p in ziparchive :
        if p.endswith(z) :            # x.zip endswith drv:/1/2/3/x.zip
            zippath = p
            newpath = ('/' if entpath != '' else '').join([zippath, entpath])
            break
    if newpath is None : print(f"[ERR:::cache] invalid decache path [{path}] ... ", file=sys.stderr); return None, None, path
    bankofpath[path] = zippath, entpath, newpath  # save cache
    return zippath, entpath, newpath

########################################

def _test(path : str) :
    # path = slashpath(path)
    zips = ['c:/a/b.zip', 'c:/1/2/3/x.zip', 'c:/1/2/3/x.z']
    encp = encache_path(zips, path)[2]
    decp = decache_path(zips, encp)[2]
    print(f"[{path}] to [{encp}] to [{decp}]")
    pass

import times
if __name__ == '__main__':
    init_cached_dir()
    stt = times.current_milli()
    _test('c:\\1\\2\\3\\x.zip\\a\\b\\c')
    _test('c:\\1\\2\\3\\x.zip\\a\\b\\c\\')
    _test('c:/1\\2//3\\x.z\\a\\b\\\\c\\\\')
    _test('c:\\1\\2\\3\\x.zip')
    _test('c:\\1\\2\\3\\x.zip\\')
    _test('c:\\1\\2\\3\\x.z\\\\')
    end = times.current_milli()
    print(f"[INF] elapsed time {(end - stt)} ms ...")
    pass

"""
[INF] zimport cache_dir ::: [W:/src.zimport.p12/.cache] from os.environ['PROJECT_HOME']
[c:\\1\\2\\3\\x.zip\\a\\b\\c] to [W:/src.zimport.p12/.cache/x.zip/a/b/c] to [c:/1/2/3/x.zip/a/b/c]
[c:\\1\\2\\3\\x.zip\\a\\b\\c\\] to [W:/src.zimport.p12/.cache/x.zip/a/b/c] to [c:/1/2/3/x.zip/a/b/c]
[c:\\1\\2\\3\\x.zip\\a\\b\\\\c\\\\] to [W:/src.zimport.p12/.cache/x.zip/a/b/c] to [c:/1/2/3/x.zip/a/b/c]
[c:\\1\\2\\3\\x.zip] to [W:/src.zimport.p12/.cache/x.zip] to [c:/1/2/3/x.zip]
[c:\\1\\2\\3\\x.zip\\] to [W:/src.zimport.p12/.cache/x.zip] to [c:/1/2/3/x.zip]
[c:\\1\\2\\3\\x.zip\\\\] to [W:/src.zimport.p12/.cache/x.zip] to [c:/1/2/3/x.zip]
[INF] elapsed time 0 ms ...
"""